from datetime import datetime
from math import inf

EIGHTEENTH = datetime(2025, 3, 23)
UNITTHRESH = {
            "second": 60, 
            "minute": 60, 
              "hour": 24, 
               "day": 7, 
              "week": 4, 
             "month": 12,
              "year": 100,
           "century": 10,
        "millennium": 1000,
     "millionennium": 1000,
     "billionennium": 1000,
    "trillionennium": inf
}

